#include "keyboardinstance.h"
#include <QFile>
#include <QProcess>

enum ListSection{
    NoSection,
    ModelSection,
    LayoutSection,
    VariantSection,
    OptionSection
};

#ifdef Q_OS_LINUX
#define XKBD_BASELIST_PATH "/usr/share/X11/xkb/rules/base.lst"
#elif defined(Q_OS_FREEBSD)
#define XKBD_BASELIST_PATH "/usr/local/share/X11/xkb/rules/base.lst"
#elif defined(Q_OS_OPENBSD)
#define XKBD_BASELIST_PATH "/usr/X11R6/share/X11/xkb/rules/base.lst"
#else
#define XKBD_BASELIST_PATH "/usr/local/share/X11/xkb/rules/base.lst"
#endif

KeyboardInstance::KeyboardInstance(QObject *parent) : QObject(parent)
{
    QByteArray configName = qgetenv("NDE_SESSION_CONFIG");
    if(configName.isEmpty())
      configName = "session";
    m_settings = new Nde::Settings(configName);

    this->loadAllKeyboardLayout();
    this->loadSelectKeyboardLayout();
}

KeyboardInstance *KeyboardInstance::getIns()
{
    static KeyboardInstance ins;
    return &ins;
}

bool KeyboardInstance::isKeyboardLayoutSelected(QString layout, QString var)
{
    for(QList<QPair<QString, QString> >::iterator it = m_selKeyboardLayouts.begin();
        it != m_selKeyboardLayouts.end(); it++)
    {
        if(it->first == layout && it->second == var)
        {
            return true;
        }
    }
    return false;
}

bool KeyboardInstance::onlyOneKeyboardLayoutSelected()
{
    return m_selKeyboardLayouts.count() <= 1;
}

void KeyboardInstance::addKeyboardLayout(QString layout, QString var)
{
    bool find = false;
    for(QList<QPair<QString, QString> >::iterator it = m_selKeyboardLayouts.begin();
        it != m_selKeyboardLayouts.end(); it++)
    {
        if(it->first == layout && it->second == var)
        {
            find = true;
        }
    }

    if(!find)
    {
        m_selKeyboardLayouts.append(QPair<QString, QString>(layout, var));
        saveKeyboardLayout();

        Q_EMIT signalSelKeyboardLayoutsChanged();
    }
}

void KeyboardInstance::removeKeyboardLayout(QString layout, QString var)
{
    for(QList<QPair<QString, QString> >::iterator it = m_selKeyboardLayouts.begin();
        it != m_selKeyboardLayouts.end(); it++)
    {
        if(it->first == layout && it->second == var)
        {
            it = m_selKeyboardLayouts.erase(it);
            saveKeyboardLayout();
            Q_EMIT signalSelKeyboardLayoutsChanged();
            break;
        }
    }
}

void KeyboardInstance::loadAllKeyboardLayout()
{
    // load known lists from xkb data files
    // XKBD_BASELIST_PATH is os dependent see keyboardlayoutconfig.h
    QFile file(QLatin1String(XKBD_BASELIST_PATH));
    if(file.open(QIODevice::ReadOnly))
    {
        ListSection section = NoSection;
        while(!file.atEnd())
        {
            QByteArray line = file.readLine().trimmed();
            if(section == NoSection)
            {
                if(line.startsWith("! model"))
                    section = ModelSection;
                else if(line.startsWith("! layout"))
                    section = LayoutSection;
                else if(line.startsWith("! variant"))
                    section = VariantSection;
                else if(line.startsWith("! option"))
                    section = OptionSection;
            }
            else
            {
                if(line.isEmpty())
                {
                    section = NoSection;
                    continue;
                }
                int sep = line.indexOf(' ');
                QString name = QString::fromLatin1(line, sep);
                while(line[sep] == ' ') // skip spaces
                    ++sep;
                QString description = QString::fromUtf8(line.constData() + sep);

                switch(section)
                {
                case LayoutSection:
                    m_allKeyboardLayouts[name] = KeyboardLayoutInfo(description);
                    break;
                case VariantSection:
                {
                    // the descriptions of variants are prefixed by their language ids
                    sep = description.indexOf(": ");
                    if(sep >= 0) {
                        QString lang = description.left(sep);
                        QMap<QString, KeyboardLayoutInfo>::iterator it = m_allKeyboardLayouts.find(lang);
                        if(it != m_allKeyboardLayouts.end()) {
                            KeyboardLayoutInfo& info = *it;
                            info.variants.append(LayoutVariantInfo(name, description.mid(sep + 2)));
                        }
                    }
                    break;
                }
                default:
                    ;
                }
            }
        }
        file.close();
    }
}

void KeyboardInstance::loadSelectKeyboardLayout()
{
    // load current settings from the output of setxkbmap command
    QProcess setxkbmap;
    setxkbmap.start(QLatin1String("setxkbmap -query -verbose 5"));
    setxkbmap.waitForFinished();
    if(setxkbmap.exitStatus() == QProcess::NormalExit) {
        QList<QByteArray> layouts, variants;
        while(!setxkbmap.atEnd()) {
            QByteArray line = setxkbmap.readLine();
            if(line.startsWith("layout:")) {
                layouts = line.mid(7).trimmed().split(',');
            }
            else if(line.startsWith("variant:")) {
                variants = line.mid(8).trimmed().split(',');
            }
        }

        const int size = layouts.size(), variantsSize = variants.size();
        for(int i = 0; i < size; ++i) {
            if(layouts.at(i).isEmpty())
                continue;
            m_selKeyboardLayouts.append(QPair<QString, QString>(layouts.at(i), variantsSize > 0 ? variants.at(i) : QString()));
        }

        setxkbmap.close();
    }
}

void KeyboardInstance::saveKeyboardLayout()
{
    //最后一个不能删除
    if(m_selKeyboardLayouts.count() <= 0) return;

    QProcess setxkbmap;
    // clear existing options
    setxkbmap.start("setxkbmap -option");
    setxkbmap.waitForFinished();
    setxkbmap.close();

    // set keyboard model
    QString command = "setxkbmap";
    command += " -model ";
    command += QString("pc105");

    // set keyboard layout
    QStringList layouts, variants;
    for(QPair<QString, QString> pair : m_selKeyboardLayouts)
    {
        layouts.append(pair.first);
        variants.append(pair.second);
    }
    command += " -layout ";
    command += layouts.join(',');
    command += " -variant ";
    command += variants.join(',');

    // execute the command line
    setxkbmap.start(command);
    setxkbmap.waitForFinished();

    // save to nde-session config file.
    m_settings->beginGroup("Keyboard");
    m_settings->setValue("layout", layouts.join(','));
    m_settings->setValue("variant", variants.join(','));
    m_settings->setValue("model", "pc105");
    m_settings->remove("options");
    m_settings->endGroup();
}
