#include "inputmethodsetting.h"
#include "ui_inputmethodsetting.h"
#include <QFile>
#include <QTextCodec>
#include <QTimer>
#include <QDebug>

SettingItem::SettingItem(FcitxIMItem *item, QFrame *parent)
    :QFrame(parent),
      select_flag(false),
      m_item(item)
{
    layout_item = new QHBoxLayout(this);
    layout_item->setContentsMargins(14, 0, 14, 0);
    layout_item->setSpacing(0);

    // label with name of input method
    label_inputmethod = new QLabel(this);
    label_inputmethod->setContentsMargins(0,0,0,0);
    QString name = item->name;
    name = name.remove(QRegExp(tr("Keyboard - ")));
    QFontMetrics fontMetrics = QFontMetrics(label_inputmethod->font());
    QString elideName = fontMetrics.elidedText(name, Qt::ElideRight, 420);
    label_inputmethod->setText(elideName);

    layout_item->addWidget(label_inputmethod);

    // spacer between input method and delete label
    horizontalSpacer = new QSpacerItem(20, 20, QSizePolicy::Expanding);
    layout_item->addSpacerItem(horizontalSpacer);

    // delete label
    label_select = new QLabel(this);
    label_select->setContentsMargins(0,0,0,0);
    label_select->setPixmap(QPixmap(":/images/select.png"));
    label_select->hide();
    label_select->installEventFilter(this);
    layout_item->addWidget(label_select);

    setFixedHeight(50);
    setStyleSheet("QFrame {border:none; border-radius:4px; background-color:#f0f1f2}");
}

SettingItem::~SettingItem(){
}

void SettingItem::updateFcitxIMItem(FcitxIMItem *item){
    m_item = item;
    QString name = item->name;
    name = name.remove(QRegExp(tr("Keyboard - ")));
    QFontMetrics fontMetrics = QFontMetrics(label_inputmethod->font());
    QString elideName = fontMetrics.elidedText(name, Qt::ElideRight, 420);
    label_inputmethod->setText(elideName);

    select_flag = m_item->enable;
    label_select->setVisible(m_item->enable);
}

void SettingItem::mousePressEvent(QMouseEvent *event){
    if(select_flag)
    {
        select_flag = false;
        label_select->hide();
    }
    else
    {
        select_flag = true;
        label_select->show();
    }
}

InputMethodSetting::InputMethodSetting(QWidget *parent) :
    Page(parent),
    ui(new Ui::InputMethodSetting)
{
    ui->setupUi(this);
    QFile f(":/qss/inputmethodsetting.qss");
    if(f.open(QIODevice::ReadOnly))
    {
        QString style = f.readAll().replace("\n", "");
        this->setStyleSheet(style);
    }

    GError *error = nullptr;
    m_improxy = fcitx_input_method_new(G_BUS_TYPE_SESSION,
                                     G_DBUS_PROXY_FLAGS_NONE,
                                     fcitx_utils_get_display_number(),
                                     nullptr,
                                     &error);
    g_signal_connect(m_improxy, "imlist-changed", G_CALLBACK(onFcitxImlistChanged), this);

    connect(ui->lineEdit_search, &QLineEdit::textChanged, this, &InputMethodSetting::onSearchTextChanged);
    updateUI();
}

InputMethodSetting::~InputMethodSetting()
{
    delete ui;
    if(m_array)
    {
        g_ptr_array_set_free_func(m_array, (GDestroyNotify) fcitx_im_item_free);
        g_ptr_array_free(m_array, FALSE);
        m_array = nullptr;
        //g_signal_handlers_disconnect_by_func(array, (gpointer)G_CALLBACK(onFcitxImlistChanged), this);
    }
    if(m_improxy)
    {
        g_signal_handlers_disconnect_by_func(m_improxy, (gpointer)G_CALLBACK(onFcitxImlistChanged), this);
        g_object_unref(m_improxy);
        m_improxy = nullptr;
    }
}

void InputMethodSetting::applyBtnClicked()
{
    for(QMap<guint, SettingItem *>::iterator it = m_settingList.begin();
        it != m_settingList.end(); it++)
    {
        SettingItem *inputitem = it.value();
        if(inputitem->isSelected())
        {
            FcitxIMItem *item = inputitem->getFcitxIMItem();
            item->enable = true;
        }
        else
        {
            FcitxIMItem *item = inputitem->getFcitxIMItem();
            item->enable = false;
        }
    }
    fcitx_input_method_set_imlist(m_improxy, m_array);
}

void InputMethodSetting::refreshPage()
{
    ui->lineEdit_search->setText("");
    updateUI();
}

bool InputMethodSetting::checkSearchFit(const QString &source, const QString &filter)
{
    QString tempFilter = filter;
    QString nospaceFilter = tempFilter.remove(QRegExp("\\s"));
    if(nospaceFilter.isEmpty()){
        return true;
    }

    QString source_string = source;
    QString pinyin_string = ConvertToQuanpin(source_string);
    if(source_string.contains(nospaceFilter, Qt::CaseInsensitive))
    {
        return true;
    }

    if(pinyin_string.contains(nospaceFilter, Qt::CaseInsensitive))
    {
        return true;
    }

    return false;
}

static bool In(int min, int max, int v)
{
    return (v >= min && v <= max);
}

static char convert(int n)
{
    /*
         * GB2312码范围 (B0A1-F7FE)
         * HiByte (B0-F7) LoByte (A1-FE)
         *
         * GBK编码范围 (8140-FEFE)
         * 包含三大部分
         * 1.汉字区
         *      a. GB2312汉字区。即GBK/2
         *      b. GB13000.1扩充汉字区。包括GBK/3(CJK汉字)和GBK/4(CJK汉字和增补汉字)
         * 2.图形符号区
         *      a. GB2312非汉字区。即GBK/1 A1A1-A9FE。还有10个小写罗马数字和GB12345增补符号
         *      b. GB13000.1扩充非汉字区。即GBK/5 A840-A9A0非汉字符号、结构符...
         * 3.用户自定义区
         *
         */
    if (In(0xB0A1,0xB0C4,n)) return 'A';
    if (In(0XB0C5,0XB2C0,n)) return 'B';
    if (In(0xB2C1,0xB4ED,n)) return 'C';
    if (In(0xB4EE,0xB6E9,n)) return 'D';
    if (In(0xB6EA,0xB7A1,n)) return 'E';
    if (In(0xB7A2,0xB8C0,n)) return 'F';
    if (In(0xB8C1,0xB9FD,n)) return 'G';
    if (In(0xB9FE,0xBBF6,n)) return 'H';
    if (In(0xBBF7,0xBFA5,n)) return 'J';
    if (In(0xBFA6,0xC0AB,n)) return 'K';
    if (In(0xC0AC,0xC2E7,n)) return 'L';
    if (In(0xC2E8,0xC4C2,n)) return 'M';
    if (In(0xC4C3,0xC5B5,n)) return 'N';
    if (In(0xC5B6,0xC5BD,n)) return 'O';
    if (In(0xC5BE,0xC6D9,n)) return 'P';
    if (In(0xC6DA,0xC8BA,n)) return 'Q';
    if (In(0xC8BB,0xC8F5,n)) return 'R';
    if (In(0xC8F6,0xCBF9,n)) return 'S';
    if (In(0xCBFA,0xCDD9,n)) return 'T';
    if (In(0xCDDA,0xCEF3,n)) return 'W';
    if (In(0xCEF4,0xD1B8,n)) return 'X';
    if (In(0xD1B9,0xD4D0,n)) return 'Y';
    if (In(0xD4D1,0xD7F9,n)) return 'Z';
    if (In(0x00, 0x7f, n)) return n;
    return '\0';
}

QString InputMethodSetting::ConvertToQuanpin(const QString zhStr)
{
    QString quanpin;
    for(int i = 0; i < zhStr.length(); i++)
    {
        QString str = zhStr.at(i);
        QTextCodec *pCodec = QTextCodec::codecForName("gb2312");
        if(!pCodec) return QChar(' ');
        QByteArray arr = pCodec->fromUnicode(str);
        wchar_t wchr = 0;

        if(arr.size() == 1)
        {
            wchr = arr.at(0) & 0xff;
        }
        else if (arr.size() == 2)
        {
            wchr = (arr.at(0) & 0xff) << 8;
            wchr |= (arr.at(1) & 0xff);
        }
        else
        {

        }

        char c = convert(wchr);
        if(c != 0)
        {
            quanpin.append(c);
        }
    }
    return quanpin;
}

void InputMethodSetting::updateUI()
{
    if(m_array != nullptr)
    {
        g_ptr_array_set_free_func(m_array, (GDestroyNotify) fcitx_im_item_free);
        g_ptr_array_free(m_array, FALSE);
        m_array = nullptr;
    }

    m_array = fcitx_input_method_get_imlist(m_improxy);
    if(nullptr == m_array)
    {
        return;
    }

    for(guint i = 0; i< m_array->len; i++)
    {
        FcitxIMItem* item = static_cast<FcitxIMItem*>(g_ptr_array_index(m_array,i));
        QMap<guint, SettingItem *>::iterator it = m_settingList.find(i);
        if(it == m_settingList.end())
        {
            SettingItem *inputItem = new SettingItem(item);
            ui->layout_list->addWidget(inputItem);
            m_settingList.insert(i, inputItem);
            inputItem->updateFcitxIMItem(item);
        }
        else {
            SettingItem * inputItem = it.value();
            inputItem->updateFcitxIMItem(item);
        }
    }
    //必须的
    onSearchTextChanged(ui->lineEdit_search->text());
}

void InputMethodSetting::onFcitxImlistChanged(FcitxInputMethod *im, InputMethodSetting* pThis)
{
    pThis->updateUI();
}

void InputMethodSetting::onSearchTextChanged(const QString &filter)
{
    for(QMap<guint, SettingItem *>::iterator it = m_settingList.begin();
        it != m_settingList.end(); it++)
    {
        SettingItem *inputitem = it.value();
        FcitxIMItem *item = inputitem->getFcitxIMItem();
        QString item_name = QString(item->name);
        if(checkSearchFit(item_name, filter))
        {
            inputitem->show();
        }
        else
        {
            inputitem->hide();
        }
    }
}
