#include "languagesetting.h"
#include "ui_languagesetting.h"
#include <QFile>
#include <QDebug>
#include <QList>
#include <QDateTime>
#include <QTextCodec>
#include <QSettings>
#include "xdgdirs.h"
#include <Settings>
#include <Nde/Settings>
#include <QStandardPaths>
#include <QDir>
#include <unistd.h>
#include "languageinstance.h"

const static QString lcLang = QStringLiteral("LANG");
const static QString lcCollate = QStringLiteral("LC_COLLATE");
const static QString lcMeasurement = QStringLiteral("LC_MEASUREMENT");
const static QString lcMonetary = QStringLiteral("LC_MONETARY");
const static QString lcNumeric = QStringLiteral("LC_NUMERIC");
const static QString lcTime = QStringLiteral("LC_TIME");

bool countryLessThan(const QLocale & c1, const QLocale & c2)
{
    return QString::localeAwareCompare(!c1.nativeCountryName().isEmpty()
                                       ? c1.nativeCountryName()
                                       : c1.countryToString(c1.country()),
                                       !c2.nativeCountryName().isEmpty()
                                       ? c2.nativeCountryName()
                                       : c2.countryToString(c2.country())) < 0;
}

//LanguageItem类初始化
LanguageItem::LanguageItem(const QLocale &locale, int id, QFrame *parent)
    :QFrame(parent),
      item_id(id),
      item_locale(locale)
{
    layout = new QHBoxLayout();
    layout->setContentsMargins(14, 9, 14, 9); //设置框内的间距
    layout->setSpacing(6);
    l_language = new QLabel(this);

    QString language_display = QString("%1 (%2)").arg(item_locale.nativeLanguageName()).arg(item_locale.nativeCountryName());
    l_language->setText(language_display);
    layout->addWidget(l_language);

    spacer = new QSpacerItem(20, 20, QSizePolicy::Expanding);
    layout->addSpacerItem(spacer);

    l_select = new QLabel(this);
    l_select->setContentsMargins(0,0,0,0);
    l_select->setPixmap(QPixmap(":/images/select.png"));
    l_select->hide();
    layout->addWidget(l_select);

    this->setFixedHeight(50);

    setLayout(layout);
}

//LanguageItem析够函数
LanguageItem::~LanguageItem(){
}

//显示或隐藏LanguageItem中l_select
void LanguageItem::ShowOrHideSelect(int on_off)
{
    if(0 == on_off)
    {
        if(!l_select->isHidden())
        {
            l_select->hide();
        }
    }
    else
    {
        l_select->show();
    }
}

//获取LanguageItem中Locale
QLocale LanguageItem::getLocale()
{
    return item_locale;
}

Languagesetting::Languagesetting(QWidget *parent) :
    Page(parent),
    ui(new Ui::Languagesetting),
    m_select_item(0),
    m_currentlocale_item(0)
{
    InitScriptCodeList();
    ui->setupUi(this);

    QVBoxLayout * conlay = new QVBoxLayout(ui->scrollAreaWidgetContents);
    conlay->setContentsMargins(0,0,0,0);
    conlay->setSpacing(2);

    QFile f(":/qss/languagesetting.qss");
    if(f.open(QIODevice::ReadOnly))
    {
        QString style = f.readAll().replace("\n", "");
        this->setStyleSheet(style);
    }

    m_currentlocale = LanguageInstance::getIns()->getCurLocale();
    QList<QLocale> allLocales = QLocale::matchingLocales(QLocale::AnyLanguage, QLocale::AnyScript, QLocale::AnyCountry);
    qSort(allLocales.begin(), allLocales.end(),countryLessThan);
    int item_id = 0;
    for(const QLocale & locale : qAsConst(allLocales))
    {
        if((!locale.nativeLanguageName().isEmpty()) && (!locale.nativeCountryName().isEmpty()))
        {
            if(locale.name() == "en_US" || locale.name() == "zh_CN")
            {
                LanguageItem *item = new LanguageItem(locale,item_id);
                if(locale == m_currentlocale)
                {
                    item->ShowOrHideSelect(1);
                    m_currentlocale_item = item_id;
                    m_select_item = item_id;
                }

                conlay->addWidget(item);
                m_languageitemlist.insert(item_id, item);
                connect(item, &LanguageItem::ShowSelect,this,&Languagesetting::ChangeSelectStatus);
                item_id++;
            }
            ui->frame_search->hide();
        }
    }
    QSpacerItem *verticalSpacer = new QSpacerItem(20, 40, QSizePolicy::Minimum, QSizePolicy::Expanding);
    conlay->addItem(verticalSpacer);

    ui->label_date_content->setText(m_currentlocale.toString(QDateTime::currentDateTime().date(),"yyyy/MM/dd"));
    ui->label_week->setText(m_currentlocale.toString(QDateTime::currentDateTime().date(),"ddd"));
    ui->label_time_content->setText(m_currentlocale.toString(QDateTime::currentDateTime().time(),"hh: mm: ss"));
    ui->label_currency_content->setText(m_currentlocale.toCurrencyString(1234.56));
    ui->label_number_content->setText(m_currentlocale.toString(1000.01));

    QString measurementSetting;
    if(m_currentlocale.measurementSystem() == QLocale::ImperialUKSystem)
    {
        measurementSetting = tr("Imperial UK");
    }
    else if (m_currentlocale.measurementSystem() == QLocale::ImperialUSSystem)
    {
        measurementSetting = tr("Imperial US");
    }
    else
    {
        measurementSetting = tr("Metric");
    }
    ui->label_measurement_content->setText(measurementSetting);

    m_interface = new QDBusInterface(QStringLiteral("org.freedesktop.Notifications"),
                                     QStringLiteral("/org/freedesktop/Notifications"),
                                     QStringLiteral("org.freedesktop.Notifications"),
                                     QDBusConnection::sessionBus(),
                                     this);

    connect(ui->lineEdit_search, &QLineEdit::textChanged, this, &Languagesetting::onSearchTextChanged);
}

Languagesetting::~Languagesetting()
{
    delete ui;
}

void Languagesetting::applyBtnClicked()
{
    if (m_currentlocale_item != m_select_item)
    {
        QList<QVariant> args;
        args << tr("Language And Input")                 //appname
             << static_cast<unsigned int>(0) //id
             << "icon-l-settings"         //icon
             << tr("Language And Input Info")                        //summary
             << tr("Please reboot to make it work.")                         //body
             << QStringList()                //actions
             << QVariantMap()                //hints
             << 5000;                     //timeout
        m_interface->callWithArgumentList(QDBus::AutoDetect, "Notify", args);

        m_currentlocale_item = m_select_item;
        LanguageItem *item = m_languageitemlist.find(m_select_item).value();
        m_currentlocale = item->getLocale();

        QString setting_value;
        Nde::Settings session_settings("session");
        session_settings.beginGroup("Environment");

        QString locale_name = m_currentlocale.name();
        if(locale_name == "az_AZ" || locale_name == "bs_BA" || locale_name == "shi_MA" ||
           locale_name == "sr_BA" || locale_name == "sr_ME" || locale_name == "sr_RS"  ||
           locale_name == "sr_XK" || locale_name == "uz_UZ" || locale_name == "vai_LR" ||
           locale_name == "zh_HK" || locale_name == "zh_MO")
        {
            int iIndex1stUnderScore = -1;
            QString code = m_ScriptCodeList.find(m_currentlocale.script()).value();
            iIndex1stUnderScore = locale_name.indexOf('_');
            setting_value = locale_name.mid(0,iIndex1stUnderScore)+'_'+code+'_'+locale_name.mid(iIndex1stUnderScore+1);
        }
        else
        {
            setting_value = locale_name;
        }

        if (!setting_value.contains('.'))
        {
            setting_value.append(QLatin1Char('.')+QTextCodec::codecForLocale()->name());
        }
        session_settings.setValue(lcLang,setting_value);
        session_settings.setValue(lcCollate,setting_value);
        session_settings.setValue(lcMeasurement,setting_value);
        session_settings.setValue(lcMonetary,setting_value);
        session_settings.setValue(lcNumeric,setting_value);
        session_settings.setValue(lcTime,setting_value);
        session_settings.endGroup();

        QString dirName;
         // WARNING: Don't use XDG_CONFIG_HOME with root because it might
         // give the user config directory if gksu-properties is set to su.
         if(geteuid() != 0) { // non-root user
             dirName = QStandardPaths::writableLocation(QStandardPaths::ConfigLocation);
         }
         if(dirName.isEmpty()) {
             dirName = QDir::homePath() % QLatin1String("/.config");
         }
        //setting environment
        QString filepath = dirName+"/environment.d/user-service.conf";
        QSettings localeSettings(filepath, QSettings::IniFormat);
        localeSettings.beginGroup("Environment");
        localeSettings.setValue(lcLang,setting_value);
        localeSettings.setValue(lcCollate,setting_value);
        localeSettings.setValue(lcMeasurement,setting_value);
        localeSettings.setValue(lcMonetary,setting_value);
        localeSettings.setValue(lcNumeric,setting_value);
        localeSettings.setValue(lcTime,setting_value);
        localeSettings.endGroup();
        localeSettings.sync();
    }
}

void Languagesetting::refreshPage()
{
    ui->lineEdit_search->setText("");
    ChangeSelectStatus(m_currentlocale_item);
}

void Languagesetting::ChangeSelectStatus(int item_id)
{
    for(int i = 0; i < m_languageitemlist.count(); i++)
    {
        LanguageItem *item = m_languageitemlist.find(i).value();
        if(i == item_id)
        {
            m_select_item = item_id;

            item->ShowOrHideSelect(1);
            QLocale preview_locale = item->getLocale();
            ui->label_date_content->setText(preview_locale.toString(QDateTime::currentDateTime().date(),"yyyy/MM/dd"));
            ui->label_week->setText(preview_locale.toString(QDateTime::currentDateTime().date(),"ddd"));
            ui->label_time_content->setText(preview_locale.toString(QDateTime::currentDateTime().time(),"hh: mm: ss"));
            ui->label_currency_content->setText(preview_locale.toCurrencyString(1234.56));
            ui->label_number_content->setText(preview_locale.toString(1000.01));

            QString measurementSetting;
            if(preview_locale.measurementSystem() == QLocale::ImperialUKSystem)
            {
                measurementSetting = tr("Imperial UK");
            }
            else if (preview_locale.measurementSystem() == QLocale::ImperialUSSystem)
            {
                measurementSetting = tr("Imperial US");
            }
            else
            {
                measurementSetting = tr("Metric");
            }
            ui->label_measurement_content->setText(measurementSetting);
        }
        else
        {
            item->ShowOrHideSelect(0);
        }
    }
}

void Languagesetting::onSearchTextChanged(const QString &filter)
{
    for(int i = 0; i < m_languageitemlist.count(); i++)
    {
        LanguageItem *item = m_languageitemlist.find(i).value();
        QLocale locale = item->getLocale();
        if(checkSearchFit(locale, filter))
        {
            item->show();
        }
        else
        {
            item->hide();
        }
    }
}

bool Languagesetting::checkSearchFit(const QLocale locale, const QString &filter)
{
    QString tempFilter = filter;
    QString nospaceFilter = tempFilter.remove(QRegExp("\\s"));
    if(nospaceFilter.isEmpty()){
        return true;
    }
    QString locale_string = locale.nativeLanguageName() + locale.nativeCountryName();
    QString pinyin_string = ConvertToQuanpin(locale_string);
    if(locale_string.contains(nospaceFilter, Qt::CaseInsensitive))
    {
        return true;
    }

    if(pinyin_string.contains(nospaceFilter, Qt::CaseInsensitive))
    {
        return true;
    }

    return false;
}

static bool In(int min, int max, int v)
{
    return (v >= min && v <= max);
}

static char convert(int n)
{
    /*
         * GB2312码范围 (B0A1-F7FE)
         * HiByte (B0-F7) LoByte (A1-FE)
         *
         * GBK编码范围 (8140-FEFE)
         * 包含三大部分
         * 1.汉字区
         *      a. GB2312汉字区。即GBK/2
         *      b. GB13000.1扩充汉字区。包括GBK/3(CJK汉字)和GBK/4(CJK汉字和增补汉字)
         * 2.图形符号区
         *      a. GB2312非汉字区。即GBK/1 A1A1-A9FE。还有10个小写罗马数字和GB12345增补符号
         *      b. GB13000.1扩充非汉字区。即GBK/5 A840-A9A0非汉字符号、结构符...
         * 3.用户自定义区
         *
         */
    if (In(0xB0A1,0xB0C4,n)) return 'A';
    if (In(0XB0C5,0XB2C0,n)) return 'B';
    if (In(0xB2C1,0xB4ED,n)) return 'C';
    if (In(0xB4EE,0xB6E9,n)) return 'D';
    if (In(0xB6EA,0xB7A1,n)) return 'E';
    if (In(0xB7A2,0xB8C0,n)) return 'F';
    if (In(0xB8C1,0xB9FD,n)) return 'G';
    if (In(0xB9FE,0xBBF6,n)) return 'H';
    if (In(0xBBF7,0xBFA5,n)) return 'J';
    if (In(0xBFA6,0xC0AB,n)) return 'K';
    if (In(0xC0AC,0xC2E7,n)) return 'L';
    if (In(0xC2E8,0xC4C2,n)) return 'M';
    if (In(0xC4C3,0xC5B5,n)) return 'N';
    if (In(0xC5B6,0xC5BD,n)) return 'O';
    if (In(0xC5BE,0xC6D9,n)) return 'P';
    if (In(0xC6DA,0xC8BA,n)) return 'Q';
    if (In(0xC8BB,0xC8F5,n)) return 'R';
    if (In(0xC8F6,0xCBF9,n)) return 'S';
    if (In(0xCBFA,0xCDD9,n)) return 'T';
    if (In(0xCDDA,0xCEF3,n)) return 'W';
    if (In(0xCEF4,0xD1B8,n)) return 'X';
    if (In(0xD1B9,0xD4D0,n)) return 'Y';
    if (In(0xD4D1,0xD7F9,n)) return 'Z';
    if (In(0x00, 0x7f, n)) return n;
    return '\0';
}

QString Languagesetting::ConvertToQuanpin(const QString zhStr)
{
    QString quanpin;
    for(int i = 0; i < zhStr.length(); i++)
    {
        QString str = zhStr.at(i);
        QTextCodec *pCodec = QTextCodec::codecForName("gb2312");
        if(!pCodec) return QChar(' ');
        QByteArray arr = pCodec->fromUnicode(str);
        wchar_t wchr = 0;

        if(arr.size() == 1)
        {
            wchr = arr.at(0) & 0xff;
        }
        else if (arr.size() == 2)
        {
            wchr = (arr.at(0) & 0xff) << 8;
            wchr |= (arr.at(1) & 0xff);
        }
        else
        {

        }

        char c = convert(wchr);
        if(c != 0)
        {
            quanpin.append(c);
        }
    }
    return quanpin;
}

void Languagesetting::InitScriptCodeList()
{
    for(unsigned int i = 0; i < script_code_list.size(); i++)
    {
        m_ScriptCodeList.insert((QLocale::Script)i,script_code_list[i]);
    }
}
